#!/usr/bin/env python
# vim: set ts=4 sts=4 sw=4 textwidth=80:

"""
Holds the PAR2Set base class
"""

__author__    = "Ira W. Snyder (devel@irasnyder.com)"
__copyright__ = "Copyright (c) 2006,2007 Ira W. Snyder (devel@irasnyder.com)"
__license__   = "GNU GPL v2 (or, at your option, any later version)"

#    Base.py
#
#    Copyright (C) 2006,2007  Ira W. Snyder (devel@irasnyder.com)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import re, os, logging
from subprocess import CalledProcessError
from PAR2Set import CompareSet, utils

# This is a generic class which handles all of the major operations that a PAR2
# set will need to be repaired, extracted, and cleaned up. For most subclasses,
# you shouldn't need to do very much.
#
# When the repair(), extract(), and delete() methods run, all of the class
# instance variables MUST match the in-filesystem state
#
# To create a new subclass:
# 1) Override the detect() method to detect your set type, raise TypeError if
#    this is not your type
# 2) Optionally, override extract() to extract your set
# 3) Override any other methods where your set differs from the Base
#    implementation

class Base(object):

    # Constructor
    # @cs an instance of CompareSet
    def __init__(self, cs, options):

        # Save the options
        self.options = options

        # The directory and parity file
        self.directory = cs.directory
        self.parityFile = cs.parityFile

        # The base name of the parity file (for matching)
        self.baseName = cs.baseName

        # Files that match by name only
        self.similarlyNamedFiles = cs.similarlyNamedFiles

        # PAR2 files from the files matched by name
        self.PAR2Files = utils.findMatches(r'^.*\.par2', self.similarlyNamedFiles)

        # All of the files protected by this set
        self.protectedFiles = cs.protectedFiles

        # Create a set of all the combined files
        self.allFiles = set(self.similarlyNamedFiles + self.protectedFiles)

        # Run the detection
        # NOTE: Python calls "down" into derived classes automatically
        # WARNING: This will raise TypeError if it doesn't match
        self.detect()

    ############################################################################

    def __str__(self):
        return '%s: %s' % (self.__class__.__name__, self.parityFile)

    ############################################################################

    def __eq__(self, rhs):
        return self.allFiles == rhs.allFiles

    ############################################################################

    # Run all operations
    def run(self):

        # Repair Stage
        try:
            self.repair()
        except (CalledProcessError, OSError):
            logging.critical('Repair stage failed for: %s' % self)
            raise

        self.updateFilesystemState()

        # Extraction Stage
        try:
            self.extract ()
        except (CalledProcessError, OSError):
            logging.critical('Extraction stage failed for: %s' % self)
            raise

        self.updateFilesystemState()

        # Deletion Stage
        try:
            self.delete ()
        except (CalledProcessError, OSError):
            logging.critical('Deletion stage failed for: %s' % self)
            raise

        logging.debug ('Successfully completed: %s' % self)

    ############################################################################

    # Run the repair
    def repair(self):

        # This is overly simple, but it works great for almost everything
        utils.runCommand(['par2repair'] + self.PAR2Files, self.directory)

    ############################################################################

    # Run the extraction
    def extract(self):
        pass

    ############################################################################

    def findDeletableFiles(self):

        regex = r'^.*\.(par2|\d|\d\d\d|rar|r\d\d|zip)$'
        files = utils.findMatches(regex, self.allFiles)

        return files

    ############################################################################

    # Run the deletion
    def delete(self):

        # If we aren't to run delete, don't
        if self.options.delete == False:
            return

        # Find all deletable files
        files = self.findDeletableFiles()
        files.sort()

        # If we are interactive, get the user's response, 'yes' or 'no'
        if self.options.interactive:
            while True:
                print 'Do you want to delete the following?:'
                for f in files:
                    print f

                s = raw_input('Delete [y/N]: ').upper()

                # If we got a valid no answer, leave now
                if s in ['N', 'NO', '']:
                    return

                # If we got a valid yes answer, delete them
                if s in ['Y', 'YES']:
                    break

                # Not a good answer, ask again
                print 'Invalid response'

        # We got a yes answer (or are non-interactive), delete the files
        for f in files:

            # Get the full filename
            fullname = os.path.join(self.directory, f)

            # Delete the file
            try:
                os.remove(fullname)
            except OSError:
                logging.error('Failed to delete: %s' % fullname)
            else:
                print 'rm', fullname
                logging.debug('Deleting: %s' % fullname)

    ############################################################################

    # Detect if the given files make up a set of this type
    #
    # Raise a TypeError if there is no match. This is meant to be called
    # from the constructor, from which we cannot return a value...
    def detect(self):

        # This must be overridden by the subclasses that actually implement
        # the functionality of rarslave

        # The original implementation used ONLY the following here:
        # self.similarlyNamedFiles
        # self.protectedFiles
        raise TypeError

    ############################################################################

    # Update the class' state to match the current filesystem state
    #
    # This must be called after any operation which can create or delete files
    # which are relevant to repair(), extract(), and delete()
    def updateFilesystemState(self):

        self.similarlyNamedFiles = utils.findFileNameMatches(self.directory,
                                                             self.baseName)
        self.allFiles = set(self.similarlyNamedFiles + self.protectedFiles)

    ############################################################################

